@@ -17,6 +17,9 @@ from __future__ import print_function
 
 import collections
 import json
+from muranopkgcheck import manager as check_manager
+from muranopkgcheck import pkg_loader as check_pkg_loader
+from muranopkgcheck import validators as check_validators
 import os
 import re
 import shutil
@@ -30,7 +33,8 @@ import zipfile
 from oslo_log import log as logging
 from oslo_serialization import jsonutils
 from oslo_utils import encodeutils
-from oslo_utils import importutils
+from oslo_utils import uuidutils
+
 import prettytable
 import requests
 import six
@@ -90,7 +94,13 @@ def print_list(objs, fields, field_labels, formatters=None, sortby=0):
                 data = getattr(o, field, None) or ''
                 row.append(data)
         pt.add_row(row)
-    print(encodeutils.safe_encode(pt.get_string()))
+
+    result = encodeutils.safe_encode(pt.get_string())
+
+    if six.PY3:
+        result = result.decode()
+
+    print(result)
 
 
 def print_dict(d, formatters=None):
@@ -104,7 +114,13 @@ def print_dict(d, formatters=None):
             pt.add_row([field, formatters[field](d[field])])
         else:
             pt.add_row([field, d[field]])
-    print(encodeutils.safe_encode(pt.get_string(sortby='Property')))
+
+    result = encodeutils.safe_encode(pt.get_string(sortby='Property'))
+
+    if six.PY3:
+        result = result.decode()
+
+    print(result)
 
 
 def find_resource(manager, name_or_id, *args, **kwargs):
@@ -149,13 +165,6 @@ def env(*vars, **kwargs):
     return kwargs.get('default', '')
 
 
-def import_versioned_module(version, submodule=None):
-    module = 'muranoclient.v%s' % version
-    if submodule:
-        module = '.'.join((module, submodule))
-    return importutils.import_module(module)
-
-
 def exit(msg=''):
     if msg:
         print(encodeutils.safe_encode(msg), file=sys.stderr)
@@ -215,6 +224,16 @@ class File(object):
         else:
             if os.path.isfile(self.name):
                 return open(self.name, mode)
+            if os.path.isdir(self.name):
+                tmp = tempfile.NamedTemporaryFile()
+                archive = zipfile.ZipFile(tmp, 'w', zipfile.ZIP_DEFLATED)
+                for root, dirs, files in os.walk(self.name):
+                    for _file in files:
+                        destination = os.path.relpath(
+                            os.path.join(root, _file), os.path.join(self.name))
+                        archive.write(os.path.join(root, _file), destination)
+                tmp.flush()
+                return open(tmp.name, mode)
             url = urllib.parse.urlparse(self.name)
             if url.scheme in ('http', 'https'):
                 resp = requests.get(self.name, stream=True)
@@ -285,7 +304,11 @@ class Package(FileWrapperMixin):
     def from_file(file_obj):
         if not isinstance(file_obj, File):
             file_obj = File(file_obj)
-        return Package(file_obj)
+        pkg = Package(file_obj)
+        errs = pkg.validate()
+        if errs:
+            raise exceptions.HTTPBadRequest(details=errs)
+        return pkg
 
     @staticmethod
     def fromFile(file_obj):
@@ -320,6 +343,16 @@ class Package(FileWrapperMixin):
             extension='.zip')
         )
 
+    def validate(self):
+        m = check_manager.Manager(self._file,
+                                  loader=check_pkg_loader.ZipLoader)
+        errors = m.validate(
+            validators=[check_validators.manifest.ManifestValidator],
+            only_errors=True)
+        if errors:
+            fmt = check_manager.PlainTextFormatter().format
+            return 'Invalid Murano package\n{}\n'.format(fmt(errors))
+
     @property
     def contents(self):
         """Contents of a package."""
@@ -368,8 +401,8 @@ class Package(FileWrapperMixin):
         if not hasattr(self, '_classes'):
             self._classes = {}
             self._resolvers = {}
-            for class_name, class_file in six.iteritems(
-                    self.manifest.get('Classes', {})):
+            for class_name, class_file in (
+                    self.manifest.get('Classes', {}).items()):
                 filename = "Classes/%s" % class_file
                 if filename not in self.contents.namelist():
                     continue
@@ -494,7 +527,7 @@ class Package(FileWrapperMixin):
     def _get_direct_deps(package, base_url, path):
         result = []
         if 'Require' in package.manifest:
-            for dep_name, ver in six.iteritems(package.manifest['Require']):
+            for dep_name, ver in package.manifest['Require'].items():
                 try:
                     req_file = Package.from_location(
                         dep_name,
@@ -554,7 +587,7 @@ def ensure_images(glance_client, image_specs, base_url,
         for key in keys:
             if key not in image:
                 LOG.warning("Image specification invalid: "
-                            "No {0} key in image ".format(key))
+                            "No {0} key in image ").format(key)
                 return False
         return True
 
@@ -738,13 +771,14 @@ def traverse_and_replace(obj,
     """Helper function that traverses object model and substitutes ids.
 
     Recursively checks values of objects found in `obj` against `pattern`,
-    and replaces strings that match pattern with uuid.uuid4(). Keeps track of
-    any replacements already made, i.e. ===id1=== would always be the same,
-    across `obj`. Uses 1st group, found in the `pattern` regexp as unique
-    identifier of a replacement
+    and replaces strings that match pattern with uuidutils.generate_uuid().
+    Keeps track of any replacements already made, i.e. ===id1=== would
+    always be the same, across `obj`. Uses 1st group, found in the `pattern`
+    regexp as unique identifier of a replacement
     """
     if replacements is None:
-        replacements = collections.defaultdict(lambda: uuid.uuid4().hex)
+        replacements = collections.defaultdict(
+            lambda: uuidutils.generate_uuid(dashed=False))
 
     def _maybe_replace(obj, key, value):
         """Check and replace value against pattern"""
@@ -752,7 +786,8 @@ def traverse_and_replace(obj,
             m = pattern.search(value)
             if m:
                 if m.group(1) not in replacements:
-                    replacements[m.group(1)] = uuid.uuid4().hex
+                    replacements[m.group(1)] = uuidutils.generate_uuid(
+                        dashed=False)
                 obj[key] = replacements[m.group(1)]
 
     if isinstance(obj, list):
