@@ -13,6 +13,7 @@
 """Application-catalog v1 package action implementation"""
 
 import collections
+import functools
 import itertools
 import os
 import shutil
@@ -25,6 +26,8 @@ from osc_lib.command import command
 from osc_lib import exceptions as exc
 from osc_lib import utils
 from oslo_log import log as logging
+from oslo_serialization import jsonutils
+from oslo_utils import strutils
 
 from muranoclient.apiclient import exceptions
 from muranoclient.common import exceptions as common_exceptions
@@ -37,6 +40,9 @@ LOG = logging.getLogger(__name__)
 
 DEFAULT_REPO_URL = "http://apps.openstack.org/api/v1/murano_repo/liberty/"
 
+_bool_from_str_strict = functools.partial(
+    strutils.bool_from_string, strict=True)
+
 
 class CreatePackage(command.Command):
     """Create an application package."""
@@ -515,3 +521,278 @@ class ImportPackage(command.Lister):
                 columns,
             ) for s in imported_list)
         )
+
+
+class ImportBundle(command.Lister):
+    """Import a bundle."""
+
+    def get_parser(self, prog_name):
+        parser = super(ImportBundle, self).get_parser(prog_name)
+        parser.add_argument(
+            'filename',
+            metavar='<FILE>',
+            nargs='+',
+            help='Bundle URL, bundle name, or path to the bundle file.'
+        )
+        parser.add_argument(
+            '--is-public',
+            action='store_true',
+            default=False,
+            help="Make the package available for users from other tenants.",
+        )
+        parser.add_argument(
+            '--exists-action',
+            default='',
+            choices=['a', 's', 'u'],
+            help='Default action when a package already exists: '
+                 '(s)kip, (u)pdate, (a)bort.'
+        )
+        parser.add_argument('--murano-repo-url',
+                            default=murano_utils.env(
+                                'MURANO_REPO_URL',
+                                default=DEFAULT_REPO_URL),
+                            help=('Defaults to env[MURANO_REPO_URL] '
+                                  'or {0}'.format(DEFAULT_REPO_URL)))
+
+        return parser
+
+    def take_action(self, parsed_args):
+
+        LOG.debug("take_action({0})".format(parsed_args))
+
+        client = self.app.client_manager.application_catalog
+
+        total_reqs = collections.OrderedDict()
+        for filename in parsed_args.filename:
+            local_path = None
+            if os.path.isfile(filename):
+                _file = filename
+                local_path = os.path.dirname(os.path.abspath(filename))
+            else:
+                print("Bundle file '{0}' does not exist, attempting "
+                      "to download".format(filename))
+                _file = murano_utils.to_url(
+                    filename,
+                    base_url=parsed_args.murano_repo_url,
+                    path='bundles/',
+                    extension='.bundle',
+                )
+
+            try:
+                bundle_file = murano_utils.Bundle.from_file(_file)
+            except Exception as e:
+                print("Failed to create bundle for '{0}', reason: {1}".format(
+                    filename, e))
+                continue
+
+            data = {"is_public": parsed_args.is_public}
+
+            for package in bundle_file.packages(
+                    base_url=parsed_args.murano_repo_url, path=local_path):
+
+                requirements = package.requirements(
+                    base_url=parsed_args.murano_repo_url,
+                    path=local_path,
+                )
+                total_reqs.update(requirements)
+
+        imported_list = []
+
+        for name, dep_package in total_reqs.items():
+            image_specs = dep_package.images()
+            if image_specs:
+                print("Inspecting required images")
+                try:
+                    imgs = parsed_args.ensure_images(
+                        glance_client=client.glance_client,
+                        image_specs=image_specs,
+                        base_url=parsed_args.murano_repo_url,
+                        local_path=local_path,
+                        is_package_public=parsed_args.is_public)
+                    for img in imgs:
+                        print("Added {0}, {1} image".format(
+                            img['name'], img['id']))
+                except Exception as e:
+                    print("Error {0} occurred while installing "
+                          "images for {1}".format(e, name))
+            try:
+                imported_package = _handle_package_exists(
+                    client, data, dep_package, parsed_args.exists_action)
+                if imported_package:
+                    imported_list.append(imported_package)
+            except exceptions.CommandError:
+                raise
+            except Exception as e:
+                print("Error {0} occurred while "
+                      "installing package {1}".format(e, name))
+
+        columns = ('id', 'name', 'fully_qualified_name', 'author', 'active',
+                   'is public', 'type', 'version')
+        column_headers = [c.capitalize() for c in columns]
+
+        return (
+            column_headers,
+            list(utils.get_item_properties(
+                s,
+                columns,
+            ) for s in imported_list)
+        )
+
+
+class ShowPackage(command.ShowOne):
+    """Display details for a package."""
+
+    def get_parser(self, prog_name):
+        parser = super(ShowPackage, self).get_parser(prog_name)
+        parser.add_argument(
+            "id",
+            metavar="<ID>",
+            help=("Package ID to show."),
+        )
+
+        return parser
+
+    def take_action(self, parsed_args):
+        LOG.debug("take_action({0})".format(parsed_args))
+        client = self.app.client_manager.application_catalog
+
+        try:
+            package = client.packages.get(parsed_args.id)
+        except common_exceptions.HTTPNotFound:
+            raise exceptions.CommandError("Package with id %s not "
+                                          "found" % parsed_args.id)
+        else:
+            to_display = dict(
+                id=package.id,
+                type=package.type,
+                owner_id=package.owner_id,
+                name=package.name,
+                fully_qualified_name=package.fully_qualified_name,
+                is_public=package.is_public,
+                enabled=package.enabled,
+                class_definitions=jsonutils.dumps(package.class_definitions,
+                                                  indent=2),
+                categories=jsonutils.dumps(package.categories, indent=2),
+                tags=jsonutils.dumps(package.tags, indent=2),
+                description=package.description
+            )
+
+        return self.dict2columns(to_display)
+
+
+class UpdatePackage(command.ShowOne):
+    """Update an existing package."""
+
+    def get_parser(self, prog_name):
+        parser = super(UpdatePackage, self).get_parser(prog_name)
+        parser.add_argument(
+            'id',
+            metavar="<ID>",
+            help="Package ID to update.",
+        )
+        parser.add_argument(
+            '--is-public',
+            type=_bool_from_str_strict,
+            metavar="{true|false}",
+            help="Make package available to users from other tenants.",
+        )
+        parser.add_argument(
+            '--enabled',
+            type=_bool_from_str_strict,
+            metavar="{true|false}",
+            help="Make package active and available for deployments.",
+        )
+        parser.add_argument(
+            '--name',
+            default=None,
+            help="New name for the package.",
+        )
+        parser.add_argument(
+            '--description',
+            default=None,
+            help="New package description.",
+        )
+        parser.add_argument(
+            '--tags',
+            metavar='<TAG>', nargs='*',
+            default=None,
+            help="A list of keywords connected to the application.",
+        )
+
+        return parser
+
+    def take_action(self, parsed_args):
+        LOG.debug("take_action({0})".format(parsed_args))
+        client = self.app.client_manager.application_catalog
+        data = {}
+        parameters = ('is_public', 'enabled',
+                      'name', 'description',
+                      'tags')
+        for parameter in parameters:
+            param_value = getattr(parsed_args, parameter, None)
+            if param_value is not None:
+                data[parameter] = param_value
+
+        _, package = client.packages.update(parsed_args.id, data)
+
+        to_display = dict(
+            id=package["id"],
+            type=package["type"],
+            owner_id=package["owner_id"],
+            name=package["name"],
+            fully_qualified_name=package["fully_qualified_name"],
+            is_public=package["is_public"],
+            enabled=package["enabled"],
+            class_definitions=jsonutils.dumps(package["class_definitions"],
+                                              indent=2),
+            categories=jsonutils.dumps(package["categories"], indent=2),
+            tags=jsonutils.dumps(package["tags"], indent=2),
+            description=package["description"]
+        )
+
+        return self.dict2columns(to_display)
+
+
+class DownloadPackage(command.Command):
+    """Download a package to a filename or stdout."""
+
+    def get_parser(self, prog_name):
+        parser = super(DownloadPackage, self).get_parser(prog_name)
+        parser.add_argument(
+            "id",
+            metavar="<ID>",
+            help=("Package ID to download."),
+        )
+        parser.add_argument(
+            "filename",
+            metavar="file", nargs="?",
+            help=("Filename to save package to. If it is not "
+                  "specified and there is no stdout redirection "
+                  "the package won't be saved."),
+        )
+
+        return parser
+
+    def take_action(self, parsed_args):
+        LOG.debug("take_action({0})".format(parsed_args))
+        client = self.app.client_manager.application_catalog
+
+        def download_to_fh(package_id, fh):
+            try:
+                fh.write(client.packages.download(package_id))
+            except common_exceptions.HTTPNotFound:
+                raise exceptions.CommandError("Package with id %s not "
+                                              "found" % parsed_args.id)
+
+        if parsed_args.filename:
+            with open(parsed_args.filename, 'wb') as fh:
+                download_to_fh(parsed_args.id, fh)
+                print("Package downloaded to %s" % parsed_args.filename)
+        elif not sys.stdout.isatty():
+            download_to_fh(parsed_args.id, sys.stdout)
+        else:
+            msg = ("No stdout redirection or local file specified for "
+                   "downloaded package. Please specify a local file to "
+                   "save downloaded package or redirect output to "
+                   "another source.")
+            raise exceptions.CommandError(msg)
